#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple, Optional, List

import json
import logging
import sys

from tap import Tap

from ..base_callback import TaskCallback
from ...networking import networkManager
from ...entities import TaskRun, secret_factory


class RemoteArgumentParser(Tap):

    refreshToken: str
    taskRunId: int
    decryptedSecrets: Optional[str]

    def configure(self) -> None:
        self.add_argument("--refreshToken", type = str)
        self.add_argument("--taskRunId", type = int)
        self.add_argument("--decryptedSecrets", nargs = "?", type = str, default = None)


def processRemote(args: Optional[List[str]] = None) -> Tuple[TaskRun, TaskCallback]:
    remoteArgumentParser, unknown = RemoteArgumentParser().parse_known_args(args)

    response = networkManager.authenticateWithRefreshToken(remoteArgumentParser.refreshToken)
    if response.hasFailed():
        raise RuntimeError(">> [Coretex] Failed to authenticate")

    taskRun: TaskRun = TaskRun.fetchById(remoteArgumentParser.taskRunId)

    if remoteArgumentParser.decryptedSecrets is not None:
        try:
            decryptedSecrets = json.loads(remoteArgumentParser.decryptedSecrets)
            if not isinstance(decryptedSecrets, dict):
                raise TypeError

            for key, value in decryptedSecrets.items():
                if isinstance(value, dict):
                    taskRun.parameters[key] = secret_factory.create(value)

                if isinstance(value, list):
                    taskRun.parameters[key] = [secret_factory.create(element) for element in value]
        except BaseException as ex:
            logging.getLogger("coretexpylib").debug(f">> [Coretex] Failed to load decrypted secrets - \"{ex}\"", exc_info = ex)

    return taskRun, TaskCallback(taskRun)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path

import json

from .base import BaseImageDataset
from ..local_dataset import LocalDataset
from ...sample import LocalImageSample
from ...annotation import ImageDatasetClass, ImageDatasetClasses


class LocalImageDataset(BaseImageDataset[LocalImageSample], LocalDataset[LocalImageSample]):  # type: ignore

    """
        Represents the Local Image Dataset class \n
        Includes functionality for working with local Image Datasets

        Properties
        ----------
        path : Path
            path to dataset
        sampleClass :
            Image Dataset Classes object
        classesPath : Path
            path to classes.json file
    """

    def __init__(self, path: Path) -> None:
        super().__init__(path, LocalImageSample)

        self.classes = ImageDatasetClasses()

        if self.classesPath.exists():
            with open(path / "classes.json") as file:
                value = json.load(file)
                self.classes = ImageDatasetClasses(
                    [ImageDatasetClass.decode(element) for element in value]
                )



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
from pathlib import Path
from zipfile import ZipFile

import logging
import shutil
import gzip

from ... import folder_manager
from ...entities import TaskRun, CustomSample, CustomDataset
from ...networking import NetworkRequestError


def createSample(name: str, dataset: CustomDataset, path: Path, taskRun: TaskRun, stepName: str, retryCount: int = 0) -> CustomSample:
    try:
        sample = dataset.add(path, name)
    except NetworkRequestError:
        if retryCount < 3:
            logging.info(f">> [Coretex] Retry count: {retryCount}")
            return createSample(name, dataset, path, taskRun, stepName, retryCount + 1)

        raise

    taskRun.createQiimeArtifact(f"{stepName}/{name}", path)

    return sample


def compressGzip(source: Path, destination: Path, deleteSource: bool = False) -> None:
    logging.info(f"{source} -> {destination}")

    with gzip.open(destination, "w") as destinationFile:
        destinationFile.write(source.read_bytes())

    if deleteSource:
        source.unlink()


def sampleNumber(sample: CustomSample) -> int:
    return int(sample.name.split("-")[0])


def isFastqMPSample(sample: CustomSample) -> bool:
    sample.unzip()

    sequenceFileNames = ["forward.fastq", "forward.fastq.gz", "sequences.fastq", "sequences.fastq.gz"]
    barcodesFileNames = ["barcodes.fastq", "barcodes.fastq.gz"]

    folderContent = list(sample.load().folderContent)

    sequenceFilePresent = any([path.name in sequenceFileNames for path in folderContent])
    barcodesFilePresent = any([path.name in barcodesFileNames for path in folderContent])

    return sequenceFilePresent and barcodesFilePresent


def isFastqDPSample(sample: CustomSample) -> bool:
    sample.unzip()

    return any([path.suffix == ".fastq" for path in sample.load().folderContent])


def isImportedSample(sample: CustomSample) -> bool:
    sample.unzip()

    sampleContent = [path.name for path in sample.load().folderContent]
    return "multiplexed-sequences.qza" in sampleContent


def isDemultiplexedSample(sample: CustomSample) -> bool:
    sample.unzip()

    sampleContent = [path.name for path in sample.load().folderContent]
    return "demux.qza" in sampleContent


def isDenoisedSample(sample: CustomSample) -> bool:
    sample.unzip()

    sampleContent = [path.name for path in sample.load().folderContent]
    return (
        "table.qza" in sampleContent and
        "rep-seqs.qza" in sampleContent and
        "stats.qza" in sampleContent
    )


def isPhylogeneticTreeSample(sample: CustomSample) -> bool:
    sample.unzip()

    sampleContent = [path.name for path in sample.load().folderContent]
    return (
        "rooted-tree.qza" in sampleContent and
        "unrooted-tree.qza" in sampleContent and
        "aligned-rep-seqs.qza" in sampleContent and
        "masked-aligned-rep-seqs.qza" in sampleContent
    )


def getFastqMPSamples(dataset: CustomDataset) -> List[CustomSample]:
    # Multiplexed fastq data
    return dataset.getSamples(isFastqMPSample)


def getFastqDPSamples(dataset: CustomDataset) -> List[CustomSample]:
    # Demultiplexed fastq data
    return dataset.getSamples(isFastqDPSample)


def getImportedSamples(dataset: CustomDataset) -> List[CustomSample]:
    return dataset.getSamples(isImportedSample)


def getDemuxSamples(dataset: CustomDataset) -> List[CustomSample]:
    return dataset.getSamples(isDemultiplexedSample)


def getDenoisedSamples(dataset: CustomDataset) -> List[CustomSample]:
    return dataset.getSamples(isDenoisedSample)


def getPhylogeneticTreeSamples(dataset: CustomDataset) -> List[CustomSample]:
    return dataset.getSamples(isPhylogeneticTreeSample)


def getMetadata(sample: CustomSample) -> Path:
    metadataPathList = list(sample.path.glob("*.tsv"))
    if len(metadataPathList) != 1:
        raise RuntimeError(f">> [Coretex] Metadata sample must contain one .tsv file. Found {len(metadataPathList)}")

    return metadataPathList[0]


def isPairedEnd(sample: CustomSample) -> bool:
    # In order to determine whether we are dealing with paired-end
    # sequences, this function unzips the qiime artifact and
    # reads the metadata, looking for the second (type) row, which will have
    # "PairedEnd" somewhere if it's paired-end

    sampleTemp = folder_manager.createTempFolder("qzaSample")
    qzaPath = list(sample.path.iterdir())[0]

    with ZipFile(qzaPath, "r") as qzaFile:
        qzaFile.extractall(sampleTemp)

    metadataPath = list(sampleTemp.rglob("*metadata.yaml"))[0]

    with metadataPath.open("r") as metadata:
        pairedEnd = "PairedEnd" in metadata.readlines()[1]

    shutil.rmtree(sampleTemp)

    return pairedEnd



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Set
from pathlib import Path

import os

from PIL import Image
from shapely.geometry import Polygon

import numpy as np
import skimage.measure

from ..base_converter import BaseConverter
from ...annotation import CoretexImageAnnotation, CoretexSegmentationInstance, ImageDatasetClass, BBox


class HumanSegmentationConverter(BaseConverter):

    def __init__(self, datasetName: str, projectId: int, datasetPath: str) -> None:
        super().__init__(datasetName, projectId, datasetPath)

        self.__imagesPath = os.path.join(datasetPath, "images")
        self.__annotationsPath = os.path.join(datasetPath, "annotations")

        self.__imageNames = list(filter(
            lambda path: path.endswith("jpg"),
            os.listdir(self.__imagesPath))
        )

    @property
    def __backgroundClass(self) -> ImageDatasetClass:
        coretexClass = self._dataset.classByName("background")
        if coretexClass is None:
            raise ValueError(f">> [Coretex] Class: (background) is not a part of dataset")

        return coretexClass

    @property
    def __humanClass(self) -> ImageDatasetClass:
        coretexClass = self._dataset.classByName("human")
        if coretexClass is None:
            raise ValueError(f">> [Coretex] Class: (human) is not a part of dataset")

        return coretexClass

    def _dataSource(self) -> List[str]:
        return self.__imageNames

    def _extractLabels(self) -> Set[str]:
        return set(["background", "human"])

    def __extractPolygons(self, annotationPath: str, imageWidth: int, imageHeight: int) -> List[List[int]]:
        maskImage = Image.open(annotationPath)
        if maskImage is None:
            return []

        maskImage = maskImage.resize((imageWidth, imageHeight), Image.Resampling.LANCZOS)
        subMaskArray = np.asarray(maskImage, dtype = np.uint8)

        # prevent segmented objects from being equal to image width/height
        subMaskArray[:, 0] = 0
        subMaskArray[0, :] = 0
        subMaskArray[:, -1] = 0
        subMaskArray[-1, :] = 0

        contours = skimage.measure.find_contours(subMaskArray, 0.5)

        segmentations: List[List[int]] = []
        for contour in contours:
            for i in range(len(contour)):
                row, col = contour[i]
                contour[i] = (col - 1, row - 1)

            # Make a polygon and simplify it
            poly = Polygon(contour)

            if poly.geom_type == 'MultiPolygon':
                # If MultiPolygon, take the smallest convex Polygon containing all the points in the object
                poly = poly.convex_hull

            # Ignore if still not a Polygon (could be a line or point)
            if poly.geom_type == 'Polygon':
                segmentation = np.array(poly.exterior.coords).ravel().tolist()
                segmentations.append(segmentation)

        # sorts polygons by size, descending
        segmentations.sort(key=len, reverse=True)

        return segmentations

    def __extractInstances(self, annotationPath: str, imageWidth: int, imageHeight: int) -> List[CoretexSegmentationInstance]:
        polygons = self.__extractPolygons(annotationPath, imageWidth, imageHeight)
        if len(polygons) == 0:
            return []

        instances: List[CoretexSegmentationInstance] = []

        largestBBox = BBox.fromPoly(polygons[0])
        firstPolygon = Polygon(np.array(polygons[0]).reshape((-1, 2)))

        backgroundBoundingBox = BBox(0, 0, imageWidth, imageHeight)
        instances.append(CoretexSegmentationInstance.create(
            self.__backgroundClass.classIds[0],
            backgroundBoundingBox,
            [backgroundBoundingBox.polygon]
        ))

        instances.append(CoretexSegmentationInstance.create(
            self.__humanClass.classIds[0],
            largestBBox,
            [polygons[0]]
        ))

        for index in range(1, len(polygons)):
            currentBBox = BBox.fromPoly(polygons[index])
            currentPolygon = Polygon(np.array(polygons[index]).reshape((-1, 2)))

            instances.append(CoretexSegmentationInstance.create(
                self.__backgroundClass.classIds[0] if currentPolygon.intersects(firstPolygon) else self.__humanClass.classIds[0],
                currentBBox,
                [polygons[index]]
            ))

        return instances

    def _extractSingleAnnotation(self, imageName: str) -> None:
        imagePath = os.path.join(self.__imagesPath, imageName)
        annotationPath = os.path.join(self.__annotationsPath, f"{Path(imagePath).stem}.png")

        image = Image.open(imagePath)
        instances = self.__extractInstances(annotationPath, image.width, image.height)

        coretexAnnotation = CoretexImageAnnotation.create(imageName, image.width, image.height, instances)
        self._saveImageAnnotationPair(imagePath, coretexAnnotation)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Type, Final


class KeyDescriptor:

    """
        Defines how json filed should be represented in python
    """

    def __init__(
        self,
        jsonName: Optional[str] = None,
        pythonType: Optional[Type] = None,
        collectionType: Optional[Type] = None,
        isEncodable: bool = True,
        isDecodable: bool = True
    ) -> None:

        self.jsonName: Final = jsonName
        self.pythonType: Final = pythonType
        self.collectionType: Final = collectionType
        self.isEncodable: Final = isEncodable
        self.isDecodable: Final = isDecodable

    def isList(self) -> bool:
        """
            Checks if descriptor represents a list or not

            Returns
            -------
            bool -> True if it does, False otherwise
        """

        return self.collectionType is not None and issubclass(self.collectionType, list)



from typing import List
from pathlib import Path

import subprocess

from threading import Thread

from .run_logger import runLogger
from ..logging import LogSeverity


def captureRunStdout(process: subprocess.Popen) -> None:
    stdout = process.stdout
    if stdout is None:
        raise ValueError("stdout is None for subprocess")

    while process.poll() is None:
        line: str = stdout.readline().decode("UTF-8")
        runLogger.logProcessOutput(line)


def captureRunStderr(process: subprocess.Popen, isEnabled: bool) -> None:
    stderr = process.stderr
    if stderr is None:
        raise ValueError("stderr is None for subprocess")

    lines: List[str] = []
    while process.poll() is None:
        line: str = stderr.readline().decode("UTF-8")
        lines.append(line)

    # Dump stderr output at the end to perserve stdout order
    for line in lines:
        if isEnabled and process.returncode == 0:
            severity = LogSeverity.warning
        elif isEnabled:
            severity = LogSeverity.fatal
        else:
            # We always want to know what gets logged to stderr
            severity = LogSeverity.debug

        runLogger.logProcessOutput(line, severity)


def executeRunLocally(
    args: List[str],
    captureErr: bool,
    cwd: Path = Path.cwd()
) -> int:

    process = subprocess.Popen(
        args,
        shell = False,
        cwd = cwd,
        stdout = subprocess.PIPE,
        stderr = subprocess.PIPE
    )

    # Run a thread which captures process stdout and prints it out to Coretex.ai console
    Thread(
        target = captureRunStdout,
        args = (process,),
        name = "Process stdout reader"
    ).start()

    if captureErr:
        # Run a thread which captures process stderr and dumps it out node log file
        Thread(
            target = captureRunStderr,
            args = (process, captureErr),
            name = "Process stderr reader"
        ).start()

    return process.wait()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Optional, Any, Union
from typing_extensions import Self
from pathlib import Path

import logging

from .base import BaseSequenceDataset
from ..network_dataset import NetworkDataset, _chunkSampleImport, _encryptedSampleImport
from ...sample import SequenceSample, CustomSample
from .... import folder_manager
from ....codable import KeyDescriptor
from ....cryptography import getProjectKey
from ....utils import file as file_utils


class SequenceDataset(BaseSequenceDataset, NetworkDataset[SequenceSample]):

    """
        Sequence Dataset class which is used for Datasets whose
        samples contain sequence data (.fasta, .fastq)
    """

    metadata: CustomSample

    def __init__(self) -> None:
        super().__init__(SequenceSample)

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["samples"] = KeyDescriptor("sessions", SequenceSample, list)

        return descriptors

    def onDecode(self) -> None:
        metadataSample = self.getSample("_metadata")
        if metadataSample is None:
            raise FileNotFoundError(">> [Coretex] _metadata sample could not be found in the dataset")

        self.metadata = CustomSample.decode(metadataSample.encode())
        self.samples = [
            sample for sample in self.samples
            if sample.id != self.metadata.id
        ]

    @classmethod
    def createSequenceDataset(
        cls,
        name: str,
        projectId: int,
        metadataPath: Union[Path, str],
        meta: Optional[Dict[str, Any]] = None
    ) -> Optional[Self]:

        """
            Creates a new sequence dataset with the provided name and metadata

            Parameters
            ----------
            name : str
                dataset name
            projectId : int
                project for which the dataset will be created
            metadataPath : Union[Path, str]
                path the zipped metadata file

            Returns
            -------
            The created sequence dataset object or None if creation failed

            Example
            -------
            >>> from coretex import SequenceDataset
            \b
            >>> dummyDataset = SequenceDataset.createSequenceDataset("dummyDataset", 123, pathToMetadata)
            >>> if dummyDataset is not None:
                    print("Dataset created successfully")
        """

        if isinstance(metadataPath, str):
            metadataPath = Path(metadataPath)

        dataset = cls.createDataset(name, projectId, meta)

        if dataset.isEncrypted:
            dataset.metadata = _encryptedSampleImport(CustomSample, "_metadata", metadataPath, dataset.id, getProjectKey(dataset.projectId))
        else:
            dataset.metadata = _chunkSampleImport(CustomSample, "_metadata", metadataPath, dataset.id)

        return dataset

    def download(self, decrypt: bool = True, ignoreCache: bool = False) -> None:
        super().download(decrypt, ignoreCache)

        self.metadata.download(decrypt, ignoreCache)

    def isPairedEnd(self) -> bool:
        """
            This function returns True if the dataset holds paired-end reads and
            False if it holds single end. Files for paired-end reads must contain
            "_R1_" and "_R2_" in their names, otherwise an exception will be raised.
            If the sample contains gzip compressed sequences, you will have to call
            Sample.unzip method first otherwise calling Sample.isPairedEnd will
            raise an exception

            Raises
            ------
            FileNotFoundError -> if no files meeting the requirements for either single-end
                or paired-end sequencing reads
            ValueError -> if dataset has a combination of single-end and paired-end samples
        """

        pairedEndSamples = [sample.isPairedEnd() for sample in self.samples]

        if all(pairedEndSamples):
            return True

        if not any(pairedEndSamples):
            return False

        raise ValueError(">> [Coretex] Dataset contains a mix of paired-end and single-end sequences. It should contain either one or the other")

    def _uploadSample(self, samplePath: Path, sampleName: str, **metadata: Any) -> SequenceSample:
        if not self._sampleType.isValidSequenceFile(samplePath):
            raise ValueError(f"\"{samplePath}\" is not a valid sequence")

        if file_utils.isArchive(samplePath):
            sample = _chunkSampleImport(self._sampleType, sampleName, samplePath, self.id)
        else:
            with folder_manager.tempFile() as archivePath:
                logging.getLogger("coretexpylib").info(f">> [Coretex] Provided Sample \"{samplePath}\" is not an archive, zipping...")
                file_utils.archive(samplePath, archivePath)

                sample = _chunkSampleImport(self._sampleType, sampleName, archivePath, self.id)

        return sample



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

import click

from ..modules import node as node_module
from ..modules.ui import clickPrompt, stdEcho, successEcho, errorEcho, previewConfig
from ..modules.update import NodeStatus, getNodeStatus, activateAutoUpdate, dumpScript, UPDATE_SCRIPT_NAME
from ..modules.utils import onBeforeCommandExecute
from ..modules.user import initializeUserSession
from ...configuration import loadConfig, saveConfig, CONFIG_DIR, isNodeConfigured
from ...utils import docker


@click.command()
@click.option("--image", type = str, help = "Docker image url")
@onBeforeCommandExecute(node_module.initializeNodeConfiguration)
def start(image: Optional[str]) -> None:
    if node_module.isRunning():
        if not clickPrompt(
            "Node is already running. Do you wish to restart the Node? (Y/n)",
            type = bool,
            default = True,
            show_default = False
        ):
            return

        node_module.stop()

    if node_module.exists():
        node_module.clean()

    config = loadConfig()

    if image is not None:
        config["image"] = image  # store forced image (flagged) so we can run autoupdate afterwards
        saveConfig(config)

    dockerImage = config["image"]

    if node_module.shouldUpdate(dockerImage):
        node_module.pull(dockerImage)

    node_module.start(dockerImage, config)
    docker.removeDanglingImages(node_module.getRepoFromImageUrl(dockerImage), node_module.getTagFromImageUrl(dockerImage))

    activateAutoUpdate(CONFIG_DIR, config)


@click.command()
def stop() -> None:
    if not node_module.isRunning():
        errorEcho("Node is already offline.")
        return

    node_module.stop()


@click.command()
@onBeforeCommandExecute(node_module.initializeNodeConfiguration)
def update() -> None:
    config = loadConfig()
    dockerImage = config["image"]

    nodeStatus = getNodeStatus()

    if nodeStatus == NodeStatus.inactive:
        errorEcho("Node is not running. To update Node you need to start it first.")
        return

    if nodeStatus == NodeStatus.reconnecting:
        errorEcho("Node is reconnecting. Cannot update now.")
        return

    if nodeStatus == NodeStatus.busy:
        if not clickPrompt(
            "Node is busy, do you wish to terminate the current execution to perform the update? (Y/n)",
            type = bool,
            default = True,
            show_default = False
        ):
            return

        node_module.stop()

    if not node_module.shouldUpdate(dockerImage):
        successEcho("Node is already up to date.")
        return

    stdEcho("Updating node...")
    node_module.pull(dockerImage)

    if getNodeStatus() == NodeStatus.busy:
        if not clickPrompt(
            "Node is busy, do you wish to terminate the current execution to perform the update? (Y/n)",
            type = bool,
            default = True,
            show_default = False
        ):
            return

    node_module.stop()

    node_module.start(dockerImage, config)
    docker.removeDanglingImages(node_module.getRepoFromImageUrl(dockerImage), node_module.getTagFromImageUrl(dockerImage))


@click.command()
@click.option("--verbose", is_flag = True, help = "Configure node settings manually.")
def config(verbose: bool) -> None:
    if node_module.isRunning():
        if not clickPrompt(
            "Node is already running. Do you wish to stop the Node? (Y/n)",
            type = bool,
            default = True,
            show_default = False
        ):
            errorEcho("If you wish to reconfigure your node, use coretex node stop commands first.")
            return

        node_module.stop()

    config = loadConfig()

    if isNodeConfigured(config):
        if not clickPrompt(
            "Node configuration already exists. Would you like to update? (Y/n)",
            type = bool,
            default = True,
            show_default = False
        ):
            return

    node_module.configureNode(config, verbose)
    saveConfig(config)
    previewConfig(config)

    # Updating auto-update script since node configuration is changed
    dumpScript(CONFIG_DIR / UPDATE_SCRIPT_NAME, config)

    successEcho("Node successfully configured.")


@click.group()
@onBeforeCommandExecute(docker.isDockerAvailable)
@onBeforeCommandExecute(initializeUserSession)
@onBeforeCommandExecute(node_module.checkResourceLimitations)
def node() -> None:
    pass


node.add_command(start, "start")
node.add_command(stop, "stop")
node.add_command(update, "update")
node.add_command(config, "config")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from .converter_processor_factory import ConverterProcessorFactory
from .base_converter import ConverterProcessorType
from ..dataset import ImageDataset


def convert(type: ConverterProcessorType, datasetName: str, projectId: int, datasetPath: str) -> Optional[ImageDataset]:
    """
        Converts and uploads the given dataset to Coretex Format

        Parameters
        ----------
        type : ConverterProcessorType
            dataset format type (coco, yolo, createML, voc, labelMe, pascalSeg)
        datasetName : str
            name of dataset
        projectId : str
            id of Coretex Project
        datasetPath : str
            path to dataset

        Returns
        -------
        Optional[ImageDataset] -> The converted ImageDataset object

        Example
        -------
        >>> from coretex import convert, ConverterProcessorType
        \b
        >>> dataset = convert(
                type = ConvertProcessorType.coco,
                datasetName = "coretex_dataset",
                projectId = 1023,
                datasetPath = "path/to/dataset"
            )
        >>> if dataset is not None:
                print("Dataset converted successfully")
    """

    return ConverterProcessorFactory(type).create(datasetName, projectId, datasetPath).convert()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, List

import logging
import traceback
import functools

from .upload_worker import LoggerUploadWorker
from ...logging import Log, LogSeverity
from ...networking import networkManager


def exceptionToString(exception: BaseException) -> str:
    tb = traceback.format_tb(exception.__traceback__)
    tb.append(str(exception))
    return "".join(tb)


def uploadTaskRunLogs(taskRunId: int, logs: List[Log]) -> bool:
    response = networkManager.post("model-queue/add-console-log", {
        "model_queue_id": taskRunId,
        "logs": [log.encode() for log in logs]
    })

    return not response.hasFailed()


class RunLogger:

    NAME = "coretex-run"

    def __init__(self) -> None:
        self._logger = logging.getLogger(RunLogger.NAME)
        self._uploadWorker: Optional[LoggerUploadWorker] = None

    def attach(self, taskRunId: int) -> None:
        if self._uploadWorker is not None:
            raise ValueError("TaskRun is already attached to logger")

        self._uploadWorker = LoggerUploadWorker(functools.partial(uploadTaskRunLogs, taskRunId))
        self._uploadWorker.start()

    def reset(self) -> None:
        if self._uploadWorker is not None:
            self._uploadWorker.stop()
            self._uploadWorker.join()

            self._uploadWorker = None

    def flushLogs(self) -> bool:
        if self._uploadWorker is None:
            return False

        return self._uploadWorker.uploadLogs()

    def logProcessOutput(self, line: str, severity: Optional[LogSeverity] = None) -> None:
        if line.strip() == "":
            return

        log, _ = Log.parse(line)

        if severity is not None:
            log.severity = severity

        self._logger.log(log.severity.getLevel(), log.message)
        if self._uploadWorker is not None:
            self._uploadWorker.add(log)

    def log(self, severity: LogSeverity, message: str, exception: Optional[BaseException] = None) -> None:
        log = Log(severity, message)
        self._logger.log(severity.getLevel(), message)

        if self._uploadWorker is not None:
            self._uploadWorker.add(log)

        if exception is not None:
            self.log(severity, exceptionToString(exception))

    def debug(self, message: str, exception: Optional[BaseException] = None) -> None:
        self.log(LogSeverity.debug, message, exception)

    def info(self, message: str, exception: Optional[BaseException] = None) -> None:
        self.log(LogSeverity.info, message, exception)

    def warning(self, message: str, exception: Optional[BaseException] = None) -> None:
        self.log(LogSeverity.warning, message, exception)

    def error(self, message: str, exception: Optional[BaseException] = None) -> None:
        self.log(LogSeverity.error, message, exception)

    def critical(self, message: str, exception: Optional[BaseException] = None) -> None:
        self.log(LogSeverity.fatal, message, exception)

    warn = warning
    fatal = critical


runLogger = RunLogger()
